/**
 * JASMINe
 * Copyright (C) 2008 Bull S.A.S.
 * Copyright (C) 2010 France Telecom
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.ow2.jasmine.monitoring.eos.monitoring;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.util.Date;

/**
 *
 * @author Philippe Durieux
 *
 */
public class ReplayFileName implements Serializable {

    /**
     * the time that the file was last modified.
     */
    private Date date;

    /**
     * Directory
     */
    private String directory = "";

    /**
     * Name of the file
     */
    private String name = "";

    /**
     * Size of the file
     */
    private long size = 0L;

    /**
     * Timestamp of the oldest record
     */
    private long dateFirst;

    /**
     * Timestamp of the newest record
     */
    private long dateLast;

    /**
     * File with header
     */
    private boolean headerType = false;

    /**
     * Index of timestamp in each line
     */
    private int timeIndex = 0;

    /**
     * File Separator, depending on the OS
     */
    private String separator;

    public ReplayFileName() {
    }

    public void setHeaderType(boolean headerType) {
        this.headerType = headerType;
    }

    public boolean isHeaderType() {
        return headerType;
    }

    public void setDate(final Date date) {
        this.date = date;
    }

    public Date getDate() {
        return date;
    }

    public long getDateFirst() {
        return dateFirst;
    }

    public void setDateFirst(long dateFirst) {
        this.dateFirst = dateFirst;
    }

    public long getDateLast() {
        return dateLast;
    }

    public void setDateLast(long dateLast) {
        this.dateLast = dateLast;
    }

    public void setDirectory(final String directory) {
        this.directory = directory;
    }

    public String getDirectory() {
        return directory;
    }

    public void setName(final String name) {
        this.name = name;
    }

    public String getName() {
        return name;
    }

    public void setSize(final long size) {
        this.size = size;
    }

    public long getSize() {
        return size;
    }

    public void setSeparator(final String separator) {
        this.separator = separator;
    }

    public String getSeparator() {
        return separator;
    }

    /**
     * Check if File is suitable for replay
     * If correct, compute the limit dates
     * @return true if suitable for replay
     */
    public boolean validate() {
        File file = new File(directory, name);
        if (! file.exists()) {
            // File does not exist
            return false;
        }
        String line = "";
        try {
            FileReader fr = new FileReader(file);
            BufferedReader buffer = new BufferedReader(fr);
            buffer.mark(500);
            line = buffer.readLine();
            if (line == null) {
                // empty file
                return false;
            }

            // Check type of file to replay
            if (line.startsWith("date")) {
                headerType = true;
                timeIndex = 1;
            } else if (line.startsWith("time")) {
                headerType = true;
                timeIndex = 0;
            } else {
                headerType = false;
                timeIndex = 0;
            }
            if (! headerType) {
                buffer.reset();
            }

            // compute timestamp limits
            long timestampFirst = Long.MAX_VALUE;
            long timestampLast = Long.MIN_VALUE;
            int nblines = 0;
            int nberrors = 0;
            while ((line = buffer.readLine()) != null) {
                String[] split = line.split(";");
                if (split.length <= 6) {
                    // Ignore line incomplete.
                    // may occur with a poll command
                    nberrors++;
                    continue;
                }
                try {
                    long timestamp = new Long(split[timeIndex]);
                    if (timestamp < timestampFirst) {
                        timestampFirst = timestamp;
                    }
                    if (timestamp > timestampLast) {
                        timestampLast = timestamp;
                    }
                    nblines++;
                } catch (NumberFormatException e) {
                    // Ignore bad line in the file.
                    // may occur if several probes use the same file
                    nberrors++;
                    continue;
                }
            }
            if (nblines == 0 || nberrors > nblines || timestampFirst > timestampLast) {
                // do not keep this file: probably not correct.
                //System.out.println("nb="+ nblines + " err=" + nberrors + " First=" + timestampFirst + "Last=" + timestampLast);
                return false;
            }
            dateFirst = timestampFirst;
            dateLast = timestampLast;
        } catch (IOException e) {
            // Cannot read file
            return false;
        }
        return true;
    }
}
