/**
 * JASMINe
 * Copyright (C) 2009 Bull S.A.S.
 * Copyright (C) 2010 France Telecom
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: MonitoringService.java 9583 2012-01-11 11:48:56Z durieuxp $
 * --------------------------------------------------------------------------
 */
package org.ow2.jasmine.monitoring.eos.monitoring.service;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.servlet.ServletException;

import org.apache.felix.ipojo.annotations.Component;
import org.apache.felix.ipojo.annotations.Instantiate;
import org.apache.felix.ipojo.annotations.Provides;
import org.apache.felix.ipojo.annotations.Requires;
import org.apache.felix.ipojo.annotations.ServiceProperty;
import org.apache.felix.ipojo.annotations.Validate;
import org.osgi.service.http.HttpContext;
import org.osgi.service.http.HttpService;
import org.osgi.service.http.NamespaceException;
import org.ow2.jasmine.event.beans.JasmineEventNotification;
import org.ow2.jasmine.monitoring.eos.common.servlet.ExportBitmapServlet;
import org.ow2.jasmine.monitoring.eos.common.servlet.UploadFileServlet;
import org.ow2.jasmine.monitoring.eos.monitoring.ConfigFile;
import org.ow2.jasmine.monitoring.eos.monitoring.EventData;
import org.ow2.jasmine.monitoring.eos.monitoring.ReplayFileName;
import org.ow2.jasmine.monitoring.eos.monitoring.Serie;
import org.ow2.jonas.lib.bootstrap.JProp;
import org.ow2.kerneos.core.service.KerneosSimpleService;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Services provided by the monitoring module.
 * @author Guillaume Renault
 * @author durieuxp
 */
@Component
@Instantiate
@Provides
public class MonitoringService implements KerneosSimpleService {

    private final Log logger = LogFactory.getLog(this.getClass());

    @ServiceProperty(name = KerneosSimpleService.ID, value = "monitoringService")
    private String id;

    /**
     * List of DashboardDataCollector, identified by their unique ident.
     */
    private Map<Integer, DashboardDataCollector> dashboardList;

    /**
     * List of files for Replay
     */
    private Set<String> replaydirlist;

    /**
     * List of directories holding configuration files (absolute path)
     */
    private Set<String> configdirlist;

    @Requires
    private HttpService httpService;

    @Requires(filter="(ID=jasmine-eos)")
    private HttpContext kerneosHttpContext;

    /**
     * count used to allocate a new uuid to a dashboard
     */
    private int dashboardIdent = 0;

    private String eostemp;

    public MonitoringService() {
    }

    @Validate
    private void start() {
        logger.debug("");
        dashboardList = new HashMap<Integer, DashboardDataCollector>();
        readConfig();
        registerServlets();
    }

    /**
     * Register servlets used by the module
     * TODO This should not be done here, because may be used by several modules.
     */
    private void registerServlets() {
        // UploadFileServlet
        UploadFileServlet s1 = new UploadFileServlet();
        try {
            httpService.registerServlet("/UploadFileServlet", s1, null, kerneosHttpContext);
        } catch (ServletException e) {
            logger.error("Cannot register UploadFileServlet: " + e);
        } catch (NamespaceException e) {
            logger.error("Cannot register UploadFileServlet: " + e);
        }
        // ExportBitmapServlet
        ExportBitmapServlet s2 = new ExportBitmapServlet();
        try {
            httpService.registerServlet("/ExportBitmapServlet", s2, null, kerneosHttpContext);
        } catch (ServletException e) {
            logger.error("Cannot register ExportBitmapServlet: " + e);
        } catch (NamespaceException e) {
            logger.error("Cannot register ExportBitmapServlet: " + e);
        }
    }

    // -----------------------------------------------------------------------------
    // Configuration Files Management
    // -----------------------------------------------------------------------------

    /**
     * Get the list of configuration files (*.xml)
     *
     * @return the list of files
     * @throws Exception If directory does not exist
     */
    public ArrayList<ConfigFile> getConfigFiles() throws Exception {
        logger.debug("");
        ArrayList<ConfigFile> result = new ArrayList<ConfigFile>();
        for (String dir : configdirlist) {
            result.addAll(getXmlFiles(dir));
        }
        return result;
    }

    /**
     * Load an xml file to configure graphs.
     *
     * @param path the xml file to load
     * @return a string representing the xml file, without the header.
     */
    public String loadConfigFile(final String path) throws Exception {
        logger.debug("Load config file: " + path);
        String result = "";
        FileReader fileReader = null;
        BufferedReader buffer = null;
        File file = new File(path);
        if (!file.exists()) {
            file = new File(JProp.getJonasBase() + File.separator + path);
        }
        if (!file.exists()) {
            throw new Exception("Impossible to locate the file: " + path);
        }
        String endofline = System.getProperty("line.separator");
        try {
            fileReader = new FileReader(file);
            buffer = new BufferedReader(fileReader);
            boolean write = false;
            String line = "";
            while ((line = buffer.readLine()) != null) {
                if (line.contains("<monitor")) {
                    write = true;
                }
                if (write) {
                    result += line + endofline;
                }
            }
            if (!write) {
                throw new Exception("Bad configuration file: No 'monitor' tag found.");
            }
            logger.debug("Config file loaded: " + result);
            return result;
        } finally {
            if (buffer != null) {
                try {
                    buffer.close();
                } catch (Exception e) {
                    logger.error("Failed closing buffer", e);
                }
            }
            if (fileReader != null) {
                try {
                    fileReader.close();
                } catch (Exception e) {
                    logger.error("Failed closing fileReader", e);
                }
            }
        }
    }

    /**
     * Save an XML config file
     *
     * @param name the name of the file
     * @param file the XML data to save
     * @return the new list of config files
     * @throws Exception
     */
    public ArrayList<ConfigFile> saveConfigFile(final String name, final String file) throws Exception {
        logger.debug("Save config file:" + name);

        // force .xml extension
        String newName = name;
        if (! newName.endsWith(".xml")) {
            newName = newName + ".xml";
        }

        // Assume the file will be put in the "monitoring" directory
        File pathname = new File(JProp.getJonasBase() + File.separator + "monitoring" + File.separator + newName);
        FileWriter f = new FileWriter(pathname);
        f.write(file);
        f.close();

        // return the new list
        return getConfigFiles();
    }

    /**
     * Remove a config file
     * @param file
     * @return
     * @throws Exception
     */
    public ArrayList<ConfigFile> removeConfigFile(final ConfigFile file) throws Exception {
        logger.debug("Remove config file:" + file.getName());

        // Remove the file
        for (String strdir : configdirlist) {
            logger.debug(strdir);
            if (strdir.endsWith(file.getDirectory())) {
                File dir = new File(strdir);
                if (!dir.exists()) {
                    logger.error("cannot find directory " + strdir);
                    break;
                }
                File f = new File(dir, file.getName());
                f.delete();
                logger.debug(file.getName() + " removed");
                break;
            }
        }

        // return the new list.
        return getConfigFiles();
    }

    /**
     * Rename a config file
     * @param file
     * @param newname
     * @return
     * @throws Exception
     */
    public ArrayList<ConfigFile> renameConfigFile(final ConfigFile file, final String newname) throws Exception {
        logger.debug("Rename config file:" + file.getName());

        // force .xml extension
        String newName = newname;
        if (! newName.endsWith(".xml")) {
            newName = newName + ".xml";
        }

        // Rename the file
        for (String strdir : configdirlist) {
            logger.debug(strdir);
            if (strdir.endsWith(file.getDirectory())) {
                File dir = new File(strdir);
                if (!dir.exists()) {
                    logger.error("cannot find " + strdir);
                    break;
                }
                File f = new File(dir, file.getName());
                File dest = new File(dir, newName);
                f.renameTo(dest);
                logger.debug(file.getName() + " renamed to " + newName);
                break;
            }
        }

        // return the new list.
        return getConfigFiles();
    }

    // -----------------------------------------------------------------------------
    // Dashboard management : Methods called from Flex client
    // -----------------------------------------------------------------------------

    /**
     * Create a new Dashboard
     *
     * @param series List of Series the dashboard is interested with
     * @return unique identifier
     * @throws Exception
     */
    public int createDashboard(final List<Serie> series, final boolean notifications) throws Exception {
        int uuid = ++dashboardIdent;
        logger.debug("dbid=" + uuid);
        DashboardDataCollector dbc = new DashboardDataCollector(uuid, series, notifications);
        dashboardList.put(uuid, dbc);
        logger.debug("db created =" + uuid);
        return uuid;
    }

    /**
     * remove a dashboard
     *
     * @param dashboard ident
     * @throws Exception
     */
    public void removeDashboard(final int dashboard) throws Exception {
        logger.debug("dbid=" + dashboard);
        DashboardDataCollector dbc = dashboardList.remove(dashboard);
        if (dbc == null) {
            throw new Exception("Bad dashboard id: " + dashboard);
        }
    }

    /**
     * Start live monitoring.
     *
     * @param dashboard ident
     * @throws Exception
     */
    public void startMonitoring(final int dashboard) throws Exception {
        logger.debug("dbid=" + dashboard);
        DashboardDataCollector dbc = dashboardList.get(dashboard);
        if (dbc == null) {
            throw new Exception("Bad dashboard id: " + dashboard);
        }
        dbc.startLiveMonitoring();
    }

    /**
     * Stop live monitoring.
     *
     * @param dashboard ident
     * @throws Exception
     */
    public void stopMonitoring(final int dashboard) throws Exception {
        logger.debug("dbid=" + dashboard);
        DashboardDataCollector dbc = dashboardList.get(dashboard);
        if (dbc == null) {
            throw new Exception("Bad dashboard id: " + dashboard);
        }
        dbc.stopLiveMonitoring();
    }

    /**
     * Get all current results for this dashboard
     *
     * @param dashboard ident
     * @return List of EventData, or null if replay is completed.
     * @throws Exception
     */
    public List<EventData> getAllResults(final int dashboard) throws Exception {
        logger.debug("dbid=" + dashboard);
        DashboardDataCollector dbc = dashboardList.get(dashboard);
        if (dbc == null) {
            logger.debug("dashboardList size = " + dashboardList.size());
            throw new Exception("Bad dashboard id: " + dashboard);
        }
        return dbc.getAllResults();
    }

    /**
     * Get all current notifications for this dashboard
     *
     * @param dashboard ident
     * @return List of notifications.
     * @throws Exception
     */
    public List<JasmineEventNotification> getAllNotifs(final int dashboard) throws Exception {
        logger.debug("dbid=" + dashboard);
        DashboardDataCollector dbc = dashboardList.get(dashboard);
        if (dbc == null) {
            logger.debug("dashboardList size = " + dashboardList.size());
            throw new Exception("Bad dashboard id: " + dashboard);
        }
        return dbc.getAllNotifs();
    }

    /**
     * Start Replay from File
     * @param dashboard
     * @param fileName
     * @param from
     * @param to
     * @throws Exception
     */
    public void startFileReplay(final int dashboard, final String fileName, final Date from, final Date to) throws Exception {
        logger.debug("dbid=" + dashboard);
        DashboardDataCollector dbc = dashboardList.get(dashboard);
        if (dbc == null) {
            throw new Exception("Bad dashboard id: " + dashboard);
        }
        dbc.startFileReplay(fileName, from, to);
    }

    /**
     * Start Replay from Database
     * @param dashboard
     * @param from
     * @param to
     * @throws Exception
     */
    public void startDatabaseReplay(final int dashboard, final Date from, final Date to) throws Exception {
        logger.debug("dbid=" + dashboard);
        DashboardDataCollector dbc = dashboardList.get(dashboard);
        if (dbc == null) {
            throw new Exception("Bad dashboard id: " + dashboard);
        }
        dbc.startDatabaseReplay(from, to);
    }

    public long getDatabaseLimits(final int dashboard) throws Exception {
        logger.debug("dbid=" + dashboard);
        DashboardDataCollector dbc = dashboardList.get(dashboard);
        if (dbc == null) {
            throw new Exception("Bad dashboard id: " + dashboard);
        }
        return dbc.getDatabaseLimits();
    }

    // -----------------------------------------------------------------------------
    // Replay from file
    // -----------------------------------------------------------------------------

    /**
     * Get the list of result files candidate to replay
     *
     * @param filter used to filter the file names in directories
     * @return the list of files
     * @throws Exception If directory does not exist
     */
    public ArrayList<ReplayFileName> getReplayFilenames(final String filter, final String filename) throws Exception {
        logger.debug(filename);

        ArrayList<ReplayFileName> result = new ArrayList<ReplayFileName>();

        // Check if a particular file is needed
        String base = null;
        if (filename.length() > 0) {
            File file = new File(filename);
            base = file.getName();
        }
        if (base != null) {
            // Case of upload
            // was: getReplayFiles(eostemp, filter)
            result.addAll(getReplayFiles(eostemp, base));
        } else {
            for (String dir : replaydirlist) {
                result.addAll(getReplayFiles(dir, filter));
            }
        }
        return result;
    }

    private ArrayList<ReplayFileName> getReplayFiles(final String directory, final String filter) {
        logger.debug(directory);
        ArrayList<ReplayFileName> result = new ArrayList<ReplayFileName>();

        // Try path as absolute, then as relative to JONAS_BASE.
        File dir = new File(directory);
        if (!dir.exists()) {
            dir = new File(JProp.getJonasBase() + File.separator + directory);
        }
        if (!dir.exists()) {
            logger.debug("cannot find directory " + directory);
            logger.debug("Nor in absolute, neither in " + JProp.getJonasBase());
            return result;
        }
        File[] files = dir.listFiles(new NameFilter(filter));
        if (files == null) {
            logger.debug("no files found in " + directory);
            return result;
        }
        for (int i = 0; i < files.length; i++) {
            ReplayFileName rfn = new ReplayFileName();
            rfn.setDirectory(dir.getPath());
            rfn.setName(files[i].getName());
            rfn.setSize(files[i].length());
            rfn.setDate(new Date(files[i].lastModified()));
            rfn.setSeparator(File.separator);
            if (rfn.validate()) {
                result.add(rfn);
                logger.debug("found " + files[i].getName());
            }
        }
        return result;
    }

    // --------------------------------------------------------------------------
    // Private functions
    // --------------------------------------------------------------------------

    /**
     * Test if the file is a Config file
     * @return
     */
    private boolean isConfigFile(final File file) {
        FileReader fileReader = null;
        BufferedReader buffer = null;
        try {
            fileReader = new FileReader(file);
            buffer = new BufferedReader(fileReader);
            String line = "";
            while ((line = buffer.readLine()) != null) {
                if (line.contains("<monitor")) {
                    return true;
                }
            }
        } catch (Exception e) {
        } finally {
            if (buffer != null) {
                try {
                    buffer.close();
                } catch (Exception e) {
                    logger.error("Failed closing buffer", e);
                }
            }
            if (fileReader != null) {
                try {
                    fileReader.close();
                } catch (Exception e) {
                    logger.error("Failed closing fileReader", e);
                }
            }
        }
        return false;
    }

    private ArrayList<ConfigFile> getXmlFiles(final String directory) {
        ArrayList<ConfigFile> result = new ArrayList<ConfigFile>();

        // Try path as absolute, then as relative to JONAS_BASE.
        File dir = new File(directory);
        if (!dir.exists()) {
            dir = new File(JProp.getJonasBase() + File.separator + directory);
        }
        if (!dir.exists()) {
            logger.error("cannot find " + directory);
            return result;
        }
        logger.debug(dir.getPath());
        File[] files = dir.listFiles(new NameFilter("*.xml"));
        if (files == null) {
            logger.warn("no files found in " + directory);
            return result;
        }
        for (int i = 0; i < files.length; i++) {
            if (isConfigFile(files[i])) {
                ConfigFile rfn = new ConfigFile();
                rfn.setDirectory(dir.getPath());
                rfn.setName(files[i].getName());
                rfn.setSize(files[i].length());
                rfn.setDate(new Date(files[i].lastModified()));
                rfn.setSeparator(File.separator);
                result.add(rfn);
                logger.debug("found " + files[i].getName());
            }
        }
        return result;
    }

    /**
     * read jasmine-monitoring.properties
     * the file may not exist.
     */
    private void readConfig() {
        JProp config = JProp.getInstance("jasmine-monitoring.properties");

        // directories for replay (usually .csv)
        replaydirlist = new HashSet<String>();
        String[] list = config.getValueAsArray("jasmine.monitoring.replaydir");
        if (list == null) {
            // Default is monitoring
            logger.debug("Missing jasmine-monitoring.properties: take default values");
            replaydirlist.add("monitoring");
        } else {
            for (int i = 0; i < list.length; i++) {
                replaydirlist.add(list[i]);
            }
        }
        // Always add the directory used for upload
        eostemp = System.getProperty("java.io.tmpdir") + File.separatorChar + "EoS.temp";
        replaydirlist.add(eostemp);

        // directories for xml configuration files
        configdirlist = new HashSet<String>();
        list = config.getValueAsArray("jasmine.monitoring.configdir");
        if (list == null) {
            // Default is monitoring
            logger.debug("No jasmine-monitoring.properties: take default values");
            configdirlist.add(JProp.getJonasBase() + File.separator + "monitoring");
        } else {
            for (int i = 0; i < list.length; i++) {
                String directory = list[i];
                // Try path as absolute, then as relative to JONAS_BASE.
                File dir = new File(directory);
                if (! dir.exists()) {
                    directory = JProp.getJonasBase() + File.separator + directory;
                    dir = new File(directory);
                }
                if (dir.exists()) {
                     configdirlist.add(directory);
                } else {
                    logger.error("cannot find " + list[i]);
                }
            }
        }
        // Always add the directory used for upload
        configdirlist.add(eostemp);
        logger.debug("jasmine-monitoring.properties read");
    }

    private boolean match(final String strmbean, final String strpattern) {
        if (strpattern == null || strpattern.length() == 0 || strpattern.equals("*:*")) {
            // No pattern defined.
            return true;
        }
        //logger.debug("compare pattern:" + strpattern + " to mbean:" + strmbean);
        ObjectName mbean = null;
        ObjectName pattern = null;
        try {
            mbean = new ObjectName(strmbean);
        } catch (MalformedObjectNameException e) {
            logger.error("Bad MBean ObjectName (MalformedObjectNameException) " + strmbean);
            return false;
        } catch (NullPointerException e) {
            logger.error("Bad MBean ObjectName (NPE) " + strmbean);
            return false;
        }
        try {
            pattern = new ObjectName(strpattern);
        } catch (MalformedObjectNameException e) {
            logger.error("Bad Pattern ObjectName (MalformedObjectNameException) " + strpattern);
            return false;
        } catch (NullPointerException e) {
            logger.error("Bad Pattern ObjectName (NPE) " + strpattern);
            return false;
        }
        if (pattern == null) {
            logger.error("Cannot get ObjectName on " + strpattern);
            return false;
        }

        return pattern.apply(mbean);
    }

}
